#include "suif_inc.h"
#include <map>
#include <sstream>
#include "dump.h"
#include "pdg.h"
#include "pdgutil.h"

#ifndef HAVE_STRNDUP
extern "C" char *strndup(const char *s, size_t n);
#endif

using pdg::PDG;
using pdg::array;
using pdg::OUnionSet;

void
dumper::visit(tree_node *t)
{
    if (t->peek_annote(k_static_root)) {
	/* Only dump a single domain */
	if (root == 0)
	    root = t->number();
	else
	    root = -1;
    }

    if (root <= 0)
	return;

    if (t->kind() != TREE_INSTR)
	return;
    instruction *in = ((tree_instr*)t)->instr();
    domain_ann *d = (domain_ann *)in->peek_annote(k_iteration_domain);
    if (d == NULL)
	return;

    node *n = new node;
    n->domain = d;
    annote_list_iter i(in->annotes());
    while (!i.is_empty()) {
	annote *an = i.step();
	if (an->name() == k_access) {
	    access_ann * acc = (access_ann *)an->data();
	    if (acc->array)
		assert(sa->vars.find(acc->array) == sa->vars.end());
	    n->accesses.push_back(acc);
	}
    }
    nodes.push_back(n);
}

dumper::~dumper()
{
    for (int i=0; i < nodes.size(); ++i)
	delete nodes[i];
}

string c_type(type_node *tn)
{
    std::ostringstream strm;

    switch (tn->op()) {
    case TYPE_INT: {
        base_type *bt = (base_type *) tn;

        if (!bt->is_signed())
	    strm << "unsigned ";
	switch (c_int_type(tn)) {
	case C_char:
	    if (bt->is_signed())
		strm << "signed ";
	    strm << "char";
	    break;
	case C_int:
	    strm << "int";
	    break;
	case C_short:
	    strm << "short";
	    break;
	case C_long:
	    strm << "long";
	    break;
	case C_longlong:
	    strm << "long long";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_FLOAT: {
	switch (c_float_type(tn)) {
	case C_float:
	    strm << "float";
	    break;
	case C_double:
	    strm << "double";
	    break;
	case C_longdouble:
	    strm << "long double";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_STRUCT: {
        struct_type *st = (struct_type *) tn;
	strm << "struct " << st->name();
	break;
    }
    default:
	assert(0);
    }

    return strm.str();
}

void dumper::dump(int argc, char * argv[])
{
    if (root <= 0)
	return;

    map<var_sym*,array*> arrays;

    PDG *pdg = new PDG;
    pdg->placement = new str(string("original"));
    pdg->root = root;

    const char *name = tp->proc()->file()->name();
    char *dot = strrchr(name, '.');
    if (!dot)
	pdg->name = new str(string(name));
    else {
	char *basename = strndup(name, dot-name);
	pdg->name = new str(string(basename));
	free(basename);
    }

    std::vector<Global_Var_Decl *> globals;
    std::vector<var_sym *> globals_sym;
    map<var_sym *, Param_Iterator *>::iterator gi;
    for (gi = sa->vars.begin(); gi != sa->vars.end(); ++gi) {
	if ((*gi).second->iterator)
	    continue;
	globals.push_back((*gi).second);
	globals_sym.push_back((*gi).first);
	pdg::parameter *p = new pdg::parameter;
	p->id = (*gi).first->sym_id();
	p->name = new str((string)(*gi).second->base_name());

	file_symtab *symtab = tp->proc()->file()->symtab();
	var_def_list_iter j(symtab->var_defs());
	while (!j.is_empty()) {
	    var_def *def = j.step();
	    if (def->variable() != (*gi).first)
		continue;
	    annote_list_iter i(def->annotes());
	    while (!i.is_empty()) {
		annote *an = i.step();
		if (an->name() != k_repeat_init)
		    continue;
		immed_list *ims = an->immeds();
		if ((*ims)[2].kind() != im_int)
		    continue;
		p->value = new integer((*ims)[2].integer());
	    }
	}

	pdg->params.push_back(p);
    }

    if (ctx) {
	pdg->context = new OUnionSet(ctx->ctx, globals);
	pdg->context->params = pdg->params;
    }

    for (int i = 0; i < nodes.size(); ++i) {
	node * n = nodes[i];
	if (!n->domain->domain->is_satisfiable())
	    continue;
	pdg::node * pn = new pdg::node;
	pdg::statement * ps = new pdg::statement;
	pdg->nodes.push_back(pn);
	pn->nr = i;
	pn->statement = ps;
	ps->operation = n->domain->insn->number();
	if (n->domain->function) {
	    ps->top_function = new pdg::function_call;
	    ps->top_function->name = new str(string(n->domain->function));
	    ps->top_function->arguments.resize(n->domain->n_args);
	}
	ps->line = source_line_num(n->domain->insn);
	pn->source = new OUnionSet(n->domain->domain, globals);
	pn->source->params = pdg->params;
	for (int j = 0; j < n->domain->plevel; ++j)
	    pn->prefix.push_back(n->domain->prefix[j]);
	int totalread = 0;
	for (int j = 0; j < n->accesses.size(); ++j)
	    totalread += !n->accesses[j]->write;
	int nread = 0;
	int nwrite = 0;
	for (int j = 0; j < n->accesses.size(); ++j) {
	    var_sym * av = n->accesses[j]->array;
	    if (av && arrays.find(av) == arrays.end()) {
		array * a = new array();
		a->id = av->sym_id();
		a->name = new str(av->name());
		type_node *tn = av->type();
		while(tn->is_array()) {
		    array_type *an = dynamic_cast<array_type*>(tn);
		    assert(an->lower_bound().is_constant());
		    assert(an->upper_bound().is_constant());
		    tn = an->elem_type();
		    a->dims.push_back(an->upper_bound().constant() -
				      an->lower_bound().constant() + 1);
		}
		a->element_type = new str(c_type(tn));
		pdg->arrays.push_back(a);
		arrays[av] = a;
	    }
	    pdg::access * a = new pdg::access;
	    a->array = av ? arrays[av] : NULL;
	    bool write = n->accesses[j]->write;
	    a->type = write ? pdg::access::write : pdg::access::read;
	    a->map = new SMap(pn->source->dim, n->accesses[j]->output,
				  n->accesses[j]->nlq, globals_sym);
	    a->map->params = pdg->params;
	    if (write)
		a->nr = totalread + nwrite++;
	    else
		a->nr = nread++;
	    ps->accesses.push_back(a);
	    if (ps->top_function && n->accesses[j]->arg_pos >= 0)
		ps->top_function->arguments[n->accesses[j]->arg_pos] = a;
	}
    }

    pdg->add_history_line("pers", argc, argv);
    pdg->Dump();
    pdg->free();
    delete pdg;
}

static void
visit (tree_node *t, void *d)
{
    ((dumper *)d)->visit(t);
}

void dumper::scan()
{
    root = 0;
    tp->map(::visit, (void *) this);
}

void dump(tree_proc *tp, context_info *ctx, int argc, char * argv[])
{
    dumper d(tp, ctx);
    d.scan();
    d.dump(argc, argv);
}
