#include <polylib/polylibgmp.h>
#include <stdio.h>  /* needed for piplib ! */
#include <piplib/piplibMP.h>
#include <assert.h>
#include "combine.h"
#include "piputil.h"

#ifndef POL_INTEGER
#define POL_INTEGER	0
#endif

#define POL_HAS(P,f)            (F_ISSET(P, f))
#define POL_SET(P,f)            (F_SET(P, f))

#define MAXRAYS    (POL_NO_DUAL | POL_INTEGER)

static PipMatrix *poly2pip(Polyhedron *P, int pos, int n, int nparam)
{
    PipMatrix *	    matrix;
    int i, j;
    unsigned int extra = P->Dimension - pos - n - nparam;

    matrix = pip_matrix_alloc(P->NbConstraints, P->Dimension+2);
    for (i = 0; i < P->NbConstraints; ++i) {
	value_assign(matrix->p[i][0], P->Constraint[i][0]);
	for (j = 0; j < n; ++j)
	    value_assign(matrix->p[i][1+j], P->Constraint[i][1+pos+j]);
	for (j = 0; j < pos; ++j)
	    value_assign(matrix->p[i][1+n+j], P->Constraint[i][1+j]);
	for (j = 0; j < extra; ++j)
	    value_assign(matrix->p[i][1+n+pos+j], P->Constraint[i][1+n+pos+j]);
	for (j = 1+pos+n+extra; j < P->Dimension+2; ++j)
	    value_assign(matrix->p[i][j], P->Constraint[i][j]);
    }

    return matrix;
}

static int max_new(PipQuast *q, int max, int d, int *maxd)
{
    PipNewparm *p;

    for (p = q->newparm; p; p = p->next)
	if (p->rank > max)
	    max = p->rank;
    if (q->condition) {
	if (++d > *maxd)
	    *maxd = d;
	max = max_new(q->next_else, max, d, maxd);
	max = max_new(q->next_then, max, d, maxd);
    }
    return max;
}

/*
 * k: the number of minimized variables that should be kept in the result
 * nv: the number of parameters in the PIP problem that should be placed
 *       before the existentially quantified variables
 * np: the number of parameters in the PIP problem that should be placed
 *         after the existentially quantified variables
 * e: number of extra existential variables that PIP proposes
 * rank: number of extra existential variables that we want to keep
 *       i.e., those that are not duplicates
 */
struct scan_data {
    Polyhedron *C;
    struct pip_domain *domain;
    int	    k;
    int	    nv;
    int	    np;
    int	    e;
    int	   *redirect;
    int	    rank;
    void    (*f)(struct scan_data *data, PipQuast *q, int i);
};

/* 
 * Return position of parameter in context that is equivalent
 * to new parameter, or -1 if there is no such parameter.
 */
static int duplicate_parm(struct scan_data *data, PipNewparm *p)
{
    int i, j, pos;
    int nparam;
    Vector *neg;
    Value tmp;
    Polyhedron *C = data->C;
    PipVector *vec = p->vector;
    Vector *v;

    if (!data->C)
	return -1;
    nparam = C->Dimension;
    for (i = nparam; i < vec->nb_elements-1; ++i)
	if (value_notzero_p(vec->the_vector[i]))
	    return -1;
    value_init(tmp);
    for (i = 0; i < C->NbConstraints; ++i) {
	pos = -1;
	for (j = 0; j < nparam; ++j) {
	    if (value_eq(C->Constraint[i][1+j], vec->the_vector[j]))
		continue;
	    if (pos != -1)
		break;
	    if (value_notzero_p(vec->the_vector[j]))
		break;
	    value_oppose(tmp, C->Constraint[i][1+j]);
	    if (value_ne(tmp, p->deno))
		break;
	    if (value_ne(C->Constraint[i][1+nparam], 
		vec->the_vector[vec->nb_elements-1]))
		    break;
	    pos = j;
	}
	if (j < nparam || pos == -1)
	    continue;
	break;
    }
    value_clear(tmp);
    if (i >= C->NbConstraints)
	return -1;
    v = Vector_Alloc(nparam+1);
    for (j = 0; j < nparam; ++j)
	value_oppose(v->p[j], vec->the_vector[j]);
    value_oppose(v->p[nparam], vec->the_vector[vec->nb_elements-1]);
    value_addto(v->p[nparam], v->p[nparam], p->deno);
    value_decrement(v->p[nparam], v->p[nparam]);
    value_assign(v->p[pos], p->deno);
    for (i = 0; i < C->NbConstraints; ++i) {
	for (j = 0; j <= nparam; ++j)
	    if (value_ne(C->Constraint[i][1+j], v->p[j]))
		break;
	if (j == nparam+1)
	    break;
    }
    Vector_Free(v);
    return i < C->NbConstraints ? pos : -1;
}

/* remove existential variables that are affine combinations of
 * other existential variables
 *
 * pos: position of first existential variable
 * n: number of existential variables
 */
static Polyhedron *remove_redundant_existentials(Polyhedron *P, int pos, int n)
{
    int i, r, s;
    Matrix M, *M2, *M3;
    Polyhedron *R;
    for (i = 0; i < n; ++i) {
	for (r = 0; r < P->NbEq; ++r)
	    if (First_Non_Zero(P->Constraint[r]+1, pos-1) == -1 &&
		value_notzero_p(P->Constraint[r][pos+i]))
		break;
	if (r == P->NbEq)
	    continue;
	if (value_notone_p(P->Constraint[r][pos+i]) &&
	    value_notmone_p(P->Constraint[r][pos+i]))
	    continue;
	if (r == P->NbEq)
	    continue;
	for (s = 0; s < P->NbConstraints; ++s) {
	    if (s == r)
		continue;
	    if (value_zero_p(P->Constraint[s][pos+i]))
		continue;
	    Combine(P->Constraint[s], P->Constraint[r],
		    P->Constraint[s], pos+i, P->Dimension+1);
	}
	/* Matrix "view" of P for RemoveColumn */
	M.NbRows = P->NbConstraints;
	M.NbColumns = P->Dimension+2;
	M.p_Init = P->p_Init;
	M.p = P->Constraint;
	M2 = RemoveColumn(&M, pos+i);
	M3 = RemoveRow(M2, r);
	Matrix_Free(M2);
	R = Constraints2Polyhedron(M3, MAXRAYS);
	Matrix_Free(M3);
	Polyhedron_Free(P);
	P = R;
	--i;
	--n;
    }
    return P;
}

static Polyhedron *D_remove_redundant_existentials(Polyhedron *D, int dim, 
						   int nparam)
{
    Polyhedron *P, *N, *R = NULL;
    for (P = D; P; P = N) {
	N = P->next;
	P->next = NULL;
	P = remove_redundant_existentials(P, 1+dim, P->Dimension-dim-nparam);
	P->next = R;
	R = P;
    }
    return R;
}

struct pip_domain {
    PipMatrix	*div;
    PipMatrix	*constraint;
    struct pip_domain	*next;
};

#define ALLOC(type) (type*)malloc(sizeof(type))

struct pip_domain* pip_domain_alloc(int rows, int vars, int divs)
{
    struct pip_domain *domain = ALLOC(struct pip_domain);
    domain->div = pip_matrix_alloc(divs, 1+vars+divs+1);
    domain->constraint = pip_matrix_alloc(rows, 1+vars+divs+1);
    domain->next = NULL;
    return domain;
}

void pip_domain_free(struct pip_domain* domain)
{
    while (domain) {
	struct pip_domain *next = domain->next;
	pip_matrix_free(domain->div);
	pip_matrix_free(domain->constraint);
	free(domain);
	domain = next;
    }
}

/* Takes redirect into account */
void copy_coefficients(struct scan_data *data, PipVector *vec, Value *dst)
{
    int j;
    int dim = data->nv+data->np+data->e;
    Vector_Copy(vec->the_vector, dst, data->nv+data->np);
    Vector_Set(dst+data->nv+data->np, 0, data->e);
    for (j = 0; j < data->rank; ++j) {
	int pos = data->redirect[j];
	value_addto(dst[pos], dst[pos], vec->the_vector[data->nv+data->np+j]);
    }
    value_assign(dst[dim], vec->the_vector[vec->nb_elements-1]);
}

/*
 * i: next row
 *
 * Dimensions in the resulting domain are:
 *	v n e (d-v-n-e)
 */
static void scan_quast_r(struct scan_data *data, PipQuast *q, int i)
{
    PipNewparm *p;
    int rank;
    int dim = data->nv+data->np+data->e;

    rank = data->rank;
    for (p = q->newparm; p; p = p->next) {
	int pos;
	PipVector *vec = p->vector;
	int np = data->np;

	pos = duplicate_parm(data, p);
	if (pos != -1) {
	    assert(pos < data->nv);
	    data->redirect[p->rank-np-data->nv] = pos;
	    continue;
	}
	data->redirect[p->rank-np-data->nv] = data->nv+data->np+data->rank;

	value_assign(data->domain->div->p[data->rank][0], p->deno);
	copy_coefficients(data, vec, data->domain->div->p[data->rank]+1);

	data->rank++;
    }

    if (q->condition) {
	int j;
	value_set_si(data->domain->constraint->p[i][0], 1);
	copy_coefficients(data, q->condition, data->domain->constraint->p[i]+1);
	scan_quast_r(data, q->next_then, i+1);

	value_set_si(data->domain->constraint->p[i][0], -1);
	Vector_Scale(data->domain->constraint->p[i]+1,
		     data->domain->constraint->p[i]+1,
		     data->domain->constraint->p[i][0], dim+1);
	value_set_si(data->domain->constraint->p[i][0], 1);
	value_decrement(data->domain->constraint->p[i][1+dim],
			data->domain->constraint->p[i][1+dim]);
	scan_quast_r(data, q->next_else, i+1);
    } else
	data->f(data, q, i);

    data->rank = rank;
}

static void scan_quast(struct scan_data *data, PipQuast *q)
{
    int		pip_param, dim;
    int		nparam, nexist, i, rows, row;
    int		d; /* maximal depth of conditions */

    pip_param = data->nv + data->np;

    d = 0;
    nexist = max_new(q, pip_param-1, 0, &d) - pip_param+1;
    if (nexist)
	data->redirect = (int *)malloc(nexist * sizeof(int));
    data->rank = 0;
    rows = d;

    /* nparam now refers to the number of parameters in the original polyhedron */
    nparam = data->np;

    data->e = nexist;
    dim = data->k+data->nv+data->np+data->e;

    data->domain = pip_domain_alloc(pip_param+d, data->nv+data->np, data->e);
    /* All parameters are/should be positive */
    for (i = 0; i < pip_param; ++i) {
	value_set_si(data->domain->constraint->p[i][0], 1);
	value_set_si(data->domain->constraint->p[i][1+i], 1);
    }
    row = pip_param;

    scan_quast_r(data, q, row);
    pip_domain_free(data->domain);
    if (nexist)
	free(data->redirect);
}

struct quast2poly_data {
    struct scan_data	scan;
    Polyhedron*		D;
    Polyhedron**	rest;
};

static void copy_constraints_2_ranges(Value **src, Value **dst, int nrow,
				      int sskip1, int dskip1, int n1,
				      int sskip2, int dskip2, int n2)
{
    int i;
    int spos2 = sskip1+n1+sskip2;
    int dpos2 = dskip1+n1+dskip2;
    for (i = 0; i < nrow; ++i) {
	value_assign(dst[i][0], src[i][0]);
	Vector_Copy(src[i]+1+sskip1, dst[i]+1+dskip1, n1);
	Vector_Copy(src[i]+1+spos2, dst[i]+1+dpos2, n2);
    }
}

static void copy_constraints_4_ranges(Value **src, Value **dst, int nrow,
				      int sskip1, int dskip1, int n1,
				      int sskip2, int dskip2, int n2,
				      int sskip3, int dskip3, int n3,
				      int sskip4, int dskip4, int n4)
{
    int i;
    int spos2 = sskip1+n1+sskip2;
    int dpos2 = dskip1+n1+dskip2;
    int spos3 = spos2+n2+sskip3;
    int dpos3 = dpos2+n2+dskip3;
    int spos4 = spos3+n3+sskip4;
    int dpos4 = dpos3+n3+dskip4;
    for (i = 0; i < nrow; ++i) {
	value_assign(dst[i][0], src[i][0]);
	Vector_Copy(src[i]+1+sskip1, dst[i]+1+dskip1, n1);
	Vector_Copy(src[i]+1+spos2, dst[i]+1+dpos2, n2);
	Vector_Copy(src[i]+1+spos3, dst[i]+1+dpos3, n3);
	Vector_Copy(src[i]+1+spos4, dst[i]+1+dpos4, n4);
    }
}

static void add_div_constraints(struct pip_domain *domain, Value **dst,
				int keep, int nv, int np, int rank)
{
    int i;
    int exist;
    int dim;

    dim = keep+nv+np+rank;
    exist = domain->div->NbColumns - 1 - nv - np - 1;
    for (i = 0; i < rank; ++i) {
	copy_constraints_4_ranges(domain->div->p+i, dst+2*i, 1,
				  0, keep, nv,
				  0, rank, np,
				  0, -np-rank, rank,
				  exist - rank, np, 1);
	/* The denominator */
	value_oppose(dst[2*i][1+keep+nv+i], domain->div->p[i][0]);
	value_set_si(dst[2*i][0], -1);
	Vector_Scale(dst[2*i]+1, dst[2*i+1]+1, dst[2*i][0], dim+1);
	value_addto(dst[2*i+1][1+dim], dst[2*i+1][1+dim], domain->div->p[i][0]);
	value_decrement(dst[2*i+1][1+dim], dst[2*i+1][1+dim]);
	value_set_si(dst[2*i][0], 1);
	value_set_si(dst[2*i+1][0], 1);
    }
}

static int copy_constraints(struct scan_data *data, Matrix *C, int row, int keep)
{
    copy_constraints_4_ranges(data->domain->constraint->p, C->p, row,
			      0, keep, data->nv,
			      0, data->rank, data->np,
			      0, -data->np-data->rank, data->rank,
			      data->e-data->rank, data->np, 1);
    add_div_constraints(data->domain, C->p+row,
			keep, data->nv, data->np, data->rank);
    row += 2*data->rank;
    if (data->C) {
	copy_constraints_2_ranges(data->C->Constraint, C->p+row,
				  data->C->NbConstraints,
				  0, keep, data->nv,
				  0, data->rank, data->np+1);
	row += data->C->NbConstraints;
    }
    return row;
}

static void add_quast_base(struct scan_data *data, PipQuast *q, int row)
{
    struct quast2poly_data *qdata = (struct quast2poly_data *)data;
    int dim = data->k+data->nv+data->np+data->rank;
    int nrow = row + 2*data->rank;

    if (q->list)
	nrow += data->k;

    if (data->C)
	nrow += data->C->NbConstraints;

    if (q->list) {
	Matrix *C;
	Polyhedron *P;

	C = Matrix_Alloc(nrow, 1+dim+1);
	row = copy_constraints(data, C, row, data->k);
	/* if data->k is zero, we are only interested in the domains
	 * where a solution exists and not in the actual solution
	 */
	if (data->k) {
	    PipList *l;
	    int j, k;
	    for (j = 0, l = q->list; j < data->k && l; ++j, l = l->next) {
		Value *v = l->vector->the_vector;

		value_set_si(C->p[row][1+j], -1);

		Vector_Copy(v, C->p[row]+1+data->k, data->nv);
		Vector_Copy(v+data->nv, C->p[row]+1+data->k+data->nv+data->rank,
			     data->np);
		for (k = data->nv+data->np; k < l->vector->nb_elements-1; ++k) {
		    int ll = data->redirect[k-data->np-data->nv];
		    if (ll < data->nv)
			ll = 1+data->k+ll;
		    else
			ll = 1+data->k+ll-data->np;
		    value_assign(C->p[row][ll], l->vector->the_vector[k]);
		}
		value_assign(C->p[row][1+dim], l->vector->the_vector[k]);

		++row;
	    }
	}
	P = Constraints2Polyhedron(C, MAXRAYS);
	Matrix_Free(C);
	P->next = qdata->D;
	qdata->D = P;
    } else if (qdata->D && qdata->rest) {  /* not interested in rest if no sol */
	Matrix *C;
	Polyhedron *P;
	C = Matrix_Alloc(nrow, 1 + dim - data->k + 1);
	row = copy_constraints(data, C, row, 0);
	P = Constraints2Polyhedron(C, MAXRAYS);
	Matrix_Free(C);
	P->next = *qdata->rest;
	*qdata->rest = P;
    }
}

/*
 * keep: the number of minimized variables that should be kept in the result
 * nvar: the number of parameters in the PIP problem that should be placed
 *       before the existentially quantified variables
 * nparam: the number of parameters in the PIP problem that should be placed
 *         after the existentially quantified variables
 * C: context of the PIP problem; may be NULL
 */
static Polyhedron *quast2poly(PipQuast *q, int keep, int nvar, int nparam,
			      Polyhedron *C, Polyhedron **rest)
{
    struct quast2poly_data	data;

    if (!q)
	return NULL;

    if (C)
	assert(C->Dimension == nvar+nparam);

    data.scan.nv = nvar;
    data.scan.k = keep;
    data.scan.np = nparam;
    data.scan.e = 0;
    data.scan.C = C;
    data.scan.redirect = NULL;

    data.D = NULL;
    data.rest = rest;
    if (rest)
	*rest = NULL;
    data.scan.f = add_quast_base;

    scan_quast(&data.scan, q);

    return data.D;
}

static void Matrix_Shift_First_Columns(Value **M, int rows, int n, int pos)
{
    int i;
    Vector *shifted = Vector_Alloc(n);
    for (i = 0; i < rows; ++i) {
	Vector_Copy(M[i]+1, shifted->p, n);
	Vector_Copy(M[i]+1+n, M[i]+1, pos);
	Vector_Copy(shifted->p, M[i]+1+pos, n);
    }
    Vector_Free(shifted);
}

/* Move first n columns to position pos, moving all columns up to pos+n-1 down
 * by n
 */
static void Polyhedron_Shift_First_Columns(Polyhedron *P, int n, int pos)
{
    int i;
    if (pos == 0)
	return;
    if (POL_HAS(P, POL_INEQUALITIES)) {
	Matrix M;
	Matrix_Shift_First_Columns(P->Constraint, P->NbConstraints, n, pos);
	/* Matrix "view" of P for Gauss */
	M.NbRows = P->NbConstraints;
	M.NbColumns = P->Dimension+2;
	M.p_Init = P->p_Init;
	M.p = P->Constraint;
	/* Put constraints back in "standard" form */
	Gauss(&M, P->NbEq, P->Dimension+1);
    }
    if (POL_HAS(P, POL_POINTS))
	Matrix_Shift_First_Columns(P->Ray, P->NbRays, n, pos);
}

Polyhedron *pip_lexminmax(Polyhedron *P, int pos, int n, int nparam, int max)
{
    PipOptions	*options;
    PipMatrix *	domain;
    unsigned int nvar = P->Dimension - nparam;
    PipMatrix   *context = pip_matrix_alloc(0, P->Dimension - n + 2);
    PipQuast	*sol;
    Polyhedron  *min;

    POL_ENSURE_INEQUALITIES(P);

    domain = poly2pip(P, pos, n, nparam);
#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, context);
#endif

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = max;
    options->Urs_unknowns = -1;
    sol = pip_solve(domain, context, -1, options);

#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    min = quast2poly(sol, n, nvar-n, nparam, NULL, NULL);
    Polyhedron_Shift_First_Columns(min, n, pos);

    pip_quast_free(sol);
    pip_matrix_free(context);
    pip_matrix_free(domain);
    pip_options_free(options);

    return min;
}

Polyhedron *Polyhedron_projectout(Polyhedron *P, int pos, int n, int nparam)
{
    PipOptions	*options;
    PipMatrix *	domain;
    unsigned int nvar = P->Dimension - nparam;
    PipMatrix   *context = pip_matrix_alloc(0, P->Dimension - n + 2);
    PipQuast	*sol;
    Polyhedron  *min;

    POL_ENSURE_INEQUALITIES(P);

    domain = poly2pip(P, pos, n, nparam);
#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, context);
#endif

    options = pip_options_init();
    options->Simplify = 1;
    options->Urs_unknowns = -1;
    sol = pip_solve(domain, context, -1, options);

#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    min = quast2poly(sol, 0, nvar - n, nparam, NULL, NULL);

    pip_quast_free(sol);
    pip_matrix_free(context);
    pip_matrix_free(domain);
    pip_options_free(options);

    return min;
}

Polyhedron *Domain_projectout(Polyhedron *D, int pos, int n, int nparam)
{
    Polyhedron *P, *R = NULL;
    for (P = D; P; P = P->next)
	R = Domain_concat(Polyhedron_projectout(P, pos, n, nparam), R);
}

static int quast_empty(PipQuast *q)
{
    return !q->newparm && !q->condition && !q->list;
}

static int return_dim(void *first, void *second)
{
    return 2*(int)first - 1;
}

/* The variables in the mapping are 
 *	previous_user next_user extra parameters
 * previous_user vars are constrained using equalities in M
 *
 * The variables in the remaining domains are
 *	first_user extra parameters
 */
struct pip_dep_result pip_inputdep(Polyhedron *D, int dim, Matrix *M)
{
    int nparam = D->Dimension - dim;
    struct pip_dep_result res;
    struct pip_dependence_info *dep;
    struct pip_rel *pr;
    Polyhedron *rel;
    Matrix *M2;
    int i;

    M2 = Matrix_Alloc(M->NbRows + D->NbConstraints, 1+dim+M->NbRows+nparam+1);
    for (i = 0; i < M->NbRows; ++i) {
	Vector_Copy(M->p[i], M2->p[i]+1, M->NbColumns-nparam-1);
	Vector_Copy(M->p[i]+M->NbColumns-nparam-1, 
		    M2->p[i]+1+dim+M->NbRows, nparam+1);
	value_set_si(M2->p[i][1+dim+i], -1);
    }
    for (i = 0; i < D->NbConstraints; ++i) {
	Vector_Copy(D->Constraint[i], M2->p[M->NbRows+i], 1+dim);
	Vector_Copy(D->Constraint[i]+1+dim, 
		    M2->p[M->NbRows+i]+1+dim+M->NbRows, nparam+1);
    }
    rel = Constraints2Polyhedron(M2, MAXRAYS);
    Matrix_Free(M2);
    dep = (struct pip_dependence_info *) 
	    malloc(sizeof(struct pip_dependence_info) + sizeof(struct pip_rel));
    dep->ctx = Domain_projectout(D, 0, dim, nparam);
    assert(dep->ctx->Dimension == nparam);
    dep->n_write = 1;
    dep->precedes_level = return_dim;
    dep->read.n_input = dim;
    dep->read.n_output = M->NbRows;
    dep->read.data = (void *)dim;
    dep->read.rel = rel;
    dep->write[0].n_input = dim;
    dep->write[0].n_output = M->NbRows;
    dep->write[0].data = (void *)dim;
    dep->write[0].rel = rel;

    pr = pip_lastwrite(dep);

    res.D = pr[1].rel;
    res.M = pr[0].rel;

    free(pr);
    Domain_Free(dep->ctx);
    Domain_Free(rel);
    free(dep);

    return res;
}

static poly2pipc(Polyhedron *P, PipMatrix *matrix,
		 int prow, int nrow, int mrow,
		 int pcol, int ncol, int mcol)
{
    int i, j;
    for (i = 0; i < nrow; ++i)
	for (j = 0; j < ncol; ++j)
	    value_assign(matrix->p[mrow+i][mcol+j], P->Constraint[prow+i][pcol+j]);
}

static sub_poly2pipc(Polyhedron *P, PipMatrix *matrix,
		     int prow, int nrow, int mrow,
		     int pcol, int ncol, int mcol)
{
    int i, j;
    for (i = 0; i < nrow; ++i)
	for (j = 0; j < ncol; ++j)
	    value_subtract(matrix->p[mrow+i][mcol+j], matrix->p[mrow+i][mcol+j], 
			   P->Constraint[prow+i][pcol+j]);
}

/* add constraints to matrix starting from row specifying that
 * the n coordinates starting at pos1 are equal to those starting at pos2
 */
static void var_equal(PipMatrix *matrix, int pos1, int n, int pos2, int row)
{
    int i;
    for (i = 0; i < n; ++i) {
	value_set_si(matrix->p[row+i][0], 0);
	value_set_si(matrix->p[row+i][pos1+i], -1);
	value_set_si(matrix->p[row+i][pos2+i], 1);
	value_set_si(matrix->p[row+i][matrix->NbColumns-1], 0);
    }
}

/* add constraints to matrix starting from row specifying that
 * the coordinate at pos1 is smaller than the one at pos2
 */
static void var_smaller(PipMatrix *matrix, int pos1, int pos2, int row)
{
    value_set_si(matrix->p[row][0], 1);
    value_set_si(matrix->p[row][pos1], -1);
    value_set_si(matrix->p[row][pos2], 1);
    value_set_si(matrix->p[row][matrix->NbColumns-1], -1);
}

/* Compute last iteration of write j that precedes a give read at level level */
static Polyhedron *last_write_2(struct pip_dependence_info *dep, 
			        Polyhedron *C, int j, int level, Polyhedron **rest)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    Polyhedron *R;
    int nrow = dep->read.rel->NbConstraints +
	       dep->write[j].rel->NbConstraints +
	       (level+1)/2;
    int nparam = dep->ctx->Dimension;
    int cextra = C->Dimension - dep->read.n_input - nparam;
    int ncol = 2 + dep->write[j].rel->Dimension + dep->read.rel->Dimension 
		 + cextra - dep->read.n_output - nparam;
    int row;
    int nout = dep->read.n_output;
    int nextra = dep->read.rel->Dimension - dep->read.n_input - nout - nparam;
    assert(nout == dep->write[j].n_output);
    assert(!dep->write[j].rel->next);
    /*
    ctx = pip_matrix_alloc(dep->ctx->NbConstraints, 
			   1 + dep->read.n_input + cextra + nparam + 1);
    poly2pipc(dep->ctx, ctx, 0, dep->ctx->NbConstraints, 0, 0, 1, 0);
    poly2pipc(dep->ctx, ctx, 0, dep->ctx->NbConstraints, 0, 
			     1, nparam+1, 1+dep->read.n_input+cextra);
    */
    ctx = poly2pip(C, 0, 0, C->Dimension);
    domain = pip_matrix_alloc(nrow, ncol);
    row = 0;
    poly2pipc(dep->write[j].rel, domain, 
	      0, dep->write[j].rel->NbConstraints, row,
	      0, 1 + dep->write[j].rel->Dimension - nparam, 0);
    poly2pipc(dep->write[j].rel, domain, 
	      0, dep->write[j].rel->NbConstraints, row,
	      1 + dep->write[j].rel->Dimension - nparam, nparam + 1,
	      domain->NbColumns - nparam - 1);
    row += dep->write[j].rel->NbConstraints;
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      0, 1, 0);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1, dep->read.n_input, 
	      domain->NbColumns - dep->read.n_input - cextra - nparam - 1);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.n_input, nout, 1 + dep->write[j].n_input);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.n_input + nout, nextra, 
	      1 + dep->write[j].rel->Dimension - nparam);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.rel->Dimension - nparam, nparam + 1,
	      domain->NbColumns - nparam - 1);
    row += dep->read.rel->NbConstraints;
    var_equal(domain, 1, level/2, 
	      domain->NbColumns - dep->read.n_input - cextra - nparam - 1, row);
    row += level/2;
    if (level % 2)
	var_smaller(domain, 1+level/2,
	  domain->NbColumns - dep->read.n_input - cextra - nparam - 1 + level/2,
	  row);
#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    R = quast2poly(sol, dep->write[j].n_input, dep->read.n_input + cextra, nparam,
			C, rest);

    pip_quast_free(sol);
    pip_options_free(options);
    pip_matrix_free(domain);
    pip_matrix_free(ctx);

    return R;
}

/* PolyLib performs Gaussian elimination using the equalities
 * in a polyhedron.  The backsubstitution may result in some
 * of the variables appearing with coefficient different from
 * one, even though we know they should have a coefficient of
 * one.
 * This procedure tries to undo some of these backsubstitutions.
 */
static void ungauss(Polyhedron *map, int row)
{
    int i, j, k;
    int dim = map->Dimension;
    Value tmp, one;
    Value **p = map->Constraint;
    value_init(tmp);
    value_init(one);
    value_set_si(one, 1);
    for (j = row+1; j < dim; ++j) {
	value_modulus(tmp, p[row][1+j], p[row][1+row]);
	if (value_zero_p(tmp))
	    continue;
	for (i = row+1; i < map->NbEq; ++i) {
	    value_modulus(tmp, p[i][1+j], p[row][1+row]);
	    if (value_zero_p(tmp))
		continue;
	    k = First_Non_Zero(p[i]+1, dim);
	    value_modulus(tmp, p[i][1+k], p[row][1+row]);
	    if (value_notzero_p(tmp))
		continue;
	    value_modulus(tmp, p[i][1+j], p[row][1+j]);
	    if (value_notzero_p(tmp))
		continue;
	    value_division(tmp, p[i][1+j], p[row][1+j]);
	    value_oppose(tmp, tmp);
	    Vector_Combine(p[row]+1, p[i]+1, p[row]+1, one, tmp, dim+1);
	    Vector_Normalize(p[row]+1, dim+1);
	    break;
	}
	if (value_one_p(p[row][row+1]))
	    break;
    }
    assert(j < dim);
    value_clear(one);
    value_clear(tmp);
}

static void assert_dependent_mapping(Polyhedron *map, int dim)
{
    int i;
    assert(map->NbConstraints >= dim);
    for (i = 0; i < dim; ++i) {
	assert(value_zero_p(map->Constraint[i][0]));
	if (!value_one_p(map->Constraint[i][1+i]))
	    ungauss(map, i);
    }
}

/* Project out first dim variables that are assumed to be
 * affine combination of the other variables.
 */
Polyhedron *Polyhedron_projectout_dep(Polyhedron *map, int dim)
{
    Matrix *M;
    Polyhedron *P;
    int i;
    assert_dependent_mapping(map, dim);
    M = Matrix_Alloc(map->NbConstraints - dim, 2 + map->Dimension - dim);
    for (i = dim; i < map->NbConstraints; ++i) {
	value_assign(M->p[i-dim][0], map->Constraint[i][0]);
	assert(First_Non_Zero(map->Constraint[i]+1, dim) == -1);
	Vector_Copy(map->Constraint[i]+1+dim, M->p[i-dim]+1, 
		    map->Dimension - dim + 1);
    }
    P = Constraints2Polyhedron(M, MAXRAYS);
    Matrix_Free(M);
    return P;
}

/* Compute last iteration of write k that precedes a given read at level read_level,
 * but follows at level write_level the iteration of write j associated
 * to the read iteration by old.
 */
static Polyhedron *last_write_3(struct pip_dependence_info *dep,
				Polyhedron *old, int j, int read_level,
				int k, int write_level, Polyhedron **rest)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    Polyhedron *R;
    int nrow = old->NbConstraints +
	       dep->read.rel->NbConstraints +
	       dep->write[k].rel->NbConstraints +
	       (read_level+1)/2 + (write_level+1)/2;
    int nparam = dep->ctx->Dimension;
    int cextra = old->Dimension - dep->write[j].n_input
				- dep->read.n_input - nparam;
    int row;
    int nout = dep->read.n_output;
    int rextra = dep->read.rel->Dimension - dep->read.n_input - nout - nparam;
    int ncol = 2 + (dep->write[k].rel->Dimension - nparam) + rextra +
		   old->Dimension;
    Polyhedron *C = Polyhedron_projectout_dep(old, dep->write[j].n_input);
    assert(nout == dep->write[j].n_output);
    assert(nout == dep->write[k].n_output);
    assert(!dep->write[j].rel->next);
    assert(!dep->write[k].rel->next);
    ctx = poly2pip(C, 0, 0, C->Dimension);

    domain = pip_matrix_alloc(nrow, ncol);
    row = 0;
    poly2pipc(old, domain, 0, old->NbConstraints, row, 0, 1, 0);
    poly2pipc(old, domain, 0, old->NbConstraints, row, 
	      1, old->Dimension+1, domain->NbColumns-old->Dimension-1);
    row += old->NbConstraints;
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      0, 1, 0);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1, dep->read.n_input, 
	      domain->NbColumns - dep->read.n_input - cextra - nparam - 1);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.n_input, nout, 1 + dep->write[k].n_input);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.n_input + nout, rextra, 
	      1 + dep->write[k].rel->Dimension - nparam);
    poly2pipc(dep->read.rel, domain,
	      0, dep->read.rel->NbConstraints, row,
	      1 + dep->read.rel->Dimension - nparam, nparam + 1,
	      domain->NbColumns - nparam - 1);
    row += dep->read.rel->NbConstraints;
    poly2pipc(dep->write[k].rel, domain, 
	      0, dep->write[k].rel->NbConstraints, row,
	      0, 1 + dep->write[k].rel->Dimension - nparam, 0);
    poly2pipc(dep->write[k].rel, domain, 
	      0, dep->write[k].rel->NbConstraints, row,
	      1 + dep->write[k].rel->Dimension - nparam, nparam + 1,
	      domain->NbColumns - nparam - 1);
    row += dep->write[k].rel->NbConstraints;

    var_equal(domain, 1, read_level/2, 
	      domain->NbColumns - dep->read.n_input - cextra - nparam - 1, row);
    row += read_level/2;
    if (read_level % 2) {
	var_smaller(domain, 1+read_level/2,
	  domain->NbColumns - dep->read.n_input - cextra - nparam - 1 + read_level/2,
	  row);
	++row;
    }

    var_equal(domain, domain->NbColumns - old->Dimension - 1, write_level/2, 1, row);
    row += write_level/2;
    if (write_level % 2) {
	var_smaller(domain, domain->NbColumns - old->Dimension - 1 + write_level/2,
	  1 + write_level/2, row);
	++row;
    }
#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    R = quast2poly(sol, dep->write[k].n_input, dep->read.n_input + cextra, nparam,
		   C, rest);

    Polyhedron_Free(C);
    pip_quast_free(sol);
    pip_options_free(options);
    pip_matrix_free(domain);
    pip_matrix_free(ctx);

    return R;
}

Polyhedron *EPoly_Intersection(Polyhedron *A, Polyhedron *B, int d, int nparam)
{
    int A_extra = A->Dimension - d - nparam;
    int B_extra = B->Dimension - d - nparam;
    Matrix *M = Matrix_Alloc(A->NbConstraints + B->NbConstraints, 
			     2 + A->Dimension + B_extra);
    Polyhedron *P;
    int i, A_row = A->NbConstraints;
    for (i = 0; i < A_row; ++i) {
	Vector_Copy(A->Constraint[i], M->p[i], 1+d+A_extra);
	Vector_Copy(A->Constraint[i]+1+d+A_extra, M->p[i]+1+d+A_extra+B_extra, 
		    nparam+1);
    }
    for (i = 0; i < B->NbConstraints; ++i) {
	Vector_Copy(B->Constraint[i], M->p[i+A_row], 1+d);
	Vector_Copy(B->Constraint[i]+1+d, M->p[i+A_row]+1+d+A_extra, 
		    B_extra+nparam+1);
    }
    P = Constraints2Polyhedron(M, MAXRAYS);
    Matrix_Free(M);
    /* return remove_redundant_existentials(P, 1+d, P->Dimension-d-nparam); */
    return P;
}

Polyhedron *EDomain_Intersection(Polyhedron *A, Polyhedron *B, int d, int nparam)
{
    Polyhedron *P, *S, *R = NULL;
    for (P = A; P; P = P->next)
	for (S = B; S; S = S->next) {
	    Polyhedron *T = EPoly_Intersection(P, S, d, nparam);
	    T->next = R;
	    R = T;
	}
    return R;
}

Polyhedron *Domain_concat(Polyhedron *head, Polyhedron *tail)
{
    Polyhedron *S;
    if (!head)
	return tail;
    for (S = head; S->next; S = S->next)
	;
    S->next = tail;
    return head;
}

/* Given a "read" access, a list of n "write" accesses and a context,
 * this function computes for each iteration of the read access,
 * the last write access of the list that wrote the data read by
 * the read access.
 * Each access is given by a relation between the loop iterators
 * and the array indices.
 * The result is a list of n+1 relations, the first n of which
 * correspond to the n write accesses and the last corresponds
 * to iterations of the read access that read data that is not
 * written by any of the n write accesses.
 *
 * The algorithm considers all levels in which a write may precede a read,
 * where a level may either be a statement level or a loop level.
 * The outermost statement level is 0, the first loop level is 1, etc...
 * The algorithm basically does the following:
 * for all levels l of the read access from innermost to outermost
 *	for all writes w that may precede the read access at that level
 *	    compute the last iteration of the write that precedes the read access
 *					    at that level
 *	    add result to possible last writes at level l of write w
 *	    for all writes w2 that we haven't considerd yet at this level that may
 *					    also precede the read access
 *		for all levels l2 of w from l to innermost
 *		    for all possible last writes dep of w at l
 *			compute last iteration of w2 between the write and read
 *								of dep
 *			add result to possible writes at level l of write w2
 *			and replace possible last write dep by remainder
 */
struct pip_rel *pip_lastwrite(struct pip_dependence_info *dep)
{
    int read_level, j, k, write_level, m;
    int readlen = 2 * dep->read.n_input + 1;
    struct pip_rel *final_rel;
    Polyhedron **temp_rel;
    Polyhedron *CA, *P, *R;
    final_rel = (struct pip_rel *)malloc((dep->n_write+1)*sizeof(struct pip_rel));

    for (j = dep->n_write-1; j >=0; --j) {
	final_rel[j].n_input = dep->write[j].n_input;
	final_rel[j].n_output = dep->read.n_input;
	final_rel[j].data = dep->write[j].data;
	final_rel[j].rel = NULL;
    }
    final_rel[dep->n_write].n_input = 0;
    final_rel[dep->n_write].n_output = dep->read.n_input;
    final_rel[dep->n_write].data = NULL;
    final_rel[dep->n_write].rel = NULL;

    assert(!dep->read.rel->next);
    CA = align_context(dep->ctx, dep->read.rel->Dimension, MAXRAYS);
    P = DomainIntersection(dep->read.rel, CA, MAXRAYS);
    Domain_Free(CA);
    R = Polyhedron_projectout(P, dep->read.n_input, 
	   dep->read.rel->Dimension - dep->read.n_input - dep->ctx->Dimension,
	   dep->ctx->Dimension);
    Domain_Free(P);
    if (!R)
	return final_rel;

    temp_rel = (Polyhedron **)malloc(dep->n_write * sizeof(Polyhedron *));

    for (read_level = readlen-1; read_level >= 0; --read_level) {
	Polyhedron *rest = NULL, *P;

	for (j = dep->n_write-1; j >=0; --j)
	    temp_rel[j] = NULL;

	for (j = dep->n_write-1; j >=0; --j) {
	    Polyhedron *P, *rest = NULL;
	    int writelen = 2 * dep->write[j].n_input + 1;
	    int plevel;

	    plevel = dep->precedes_level(dep->write[j].data, dep->read.data);
	    if (plevel < read_level)
		continue;
	    if (!(read_level % 2) && plevel > read_level)
		continue;

	    for (P = R; P; P = P->next) {
		Polyhedron *T, *trest;

		T = last_write_2(dep, P, j, read_level, &trest);
		if (!T) {
		    rest = Domain_concat(Polyhedron_Copy(P), rest);
		    continue;
		}
		rest = Domain_concat(trest, rest);
		T = D_remove_redundant_existentials(T, 
			    final_rel[j].n_input + final_rel[j].n_output,
			    dep->ctx->Dimension);
		temp_rel[j] = Domain_concat(T, temp_rel[j]);
	    }

	    /* the middle write */
	    for (k = j-1; k >= 0; --k) {
		int plevel, plevel2;
		plevel = dep->precedes_level(dep->write[k].data, dep->read.data);
		if (plevel < read_level)
		    continue;
		if (!(read_level % 2) && plevel > read_level)
		    continue;

		plevel2 = dep->precedes_level(dep->write[j].data,
						dep->write[k].data);

		for (write_level = read_level; write_level < writelen; ++write_level) {
		    Polyhedron *T;

		    if (plevel2 < write_level)
			continue;
		    if (!(write_level % 2) && plevel2 > write_level)
			continue;

		    /* the three write thing */
		    for (P = temp_rel[j], T = NULL; P; P = P->next) {
			Polyhedron *T2, *trest;
			T2 = last_write_3(dep, P, j, read_level, k, write_level,
					  &trest);
			if (!T2) {
			    T = Domain_concat(Polyhedron_Copy(P), T);
			    continue;
			}
			if (trest) {
			    Polyhedron *RA, *next, *I;
			    RA = align_context(trest, trest->Dimension + 
					       dep->write[j].n_input, MAXRAYS);
			    next = P->next;
			    P->next = NULL;
			    I = EDomain_Intersection(P, RA,
				    dep->write[j].n_input + dep->read.n_input,
				    dep->ctx->Dimension);
			    P->next = next;
			    Domain_Free(RA);
			    Domain_Free(trest);
			    T = Domain_concat(I, T);
			}
			T2 = D_remove_redundant_existentials(T2, 
				    final_rel[k].n_input + final_rel[k].n_output,
				    dep->ctx->Dimension);
			temp_rel[k] = Domain_concat(T2, temp_rel[k]);
		    }
		    Domain_Free(temp_rel[j]);
		    temp_rel[j] = T;
		}
	    }

	    Domain_Free(R);
	    R = rest;
	    while(temp_rel[j]) {
		Polyhedron *next = temp_rel[j]->next;
		temp_rel[j]->next = final_rel[j].rel;
		final_rel[j].rel = temp_rel[j];
		temp_rel[j] = next;
	    }
	}
    }
    final_rel[dep->n_write].rel = R;
    free(temp_rel);
    return final_rel;
}

static int copy_domain_constraints(Polyhedron *dep1, Polyhedron *dep2,
				   int n_write1, int n_write2,
				   int n_read, Polyhedron *C, 
				   PipMatrix *domain, int row, int col)
{
    int nparam = C->Dimension;
    int extra1 = dep1->Dimension - n_write1 - n_read - nparam;
    int extra2 = dep2->Dimension - n_write2 - n_read - nparam;

    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row, 
	       0, 1, 0);
    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row,
	      1+n_write1, n_read + extra1, 1+col);
    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row,
	      1+n_write1+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    row += dep1->NbConstraints - n_write1;
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	       0, 1, 0);
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	      1+n_write2, n_read + extra2, 1+col+n_read+extra1);
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	      1+n_write2+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += dep2->NbConstraints - n_write2;
    poly2pipc(C, domain, 0, C->NbConstraints, row, 0, 1, 0);
    poly2pipc(C, domain, 0, C->NbConstraints, row, 
	      1, nparam+1, domain->NbColumns-nparam-1);
    row += C->NbConstraints;
    return row;
}

/* only use on result of maximization problems */
static int lex_sign(PipList *list, int dim)
{
    int i;
    for (i = 0; i < dim; ++i) {
	assert(list);
	if (value_zero_p(list->vector->the_deno[0]))
	    return 1;
	if (value_neg_p(list->vector->the_vector[0]))
	    return -1;
	if (value_pos_p(list->vector->the_vector[0]))
	    return 1;
	list = list->next;
    }
    return 0;
}

/*
 * Check whether there is any reordering between dep1 and dep2.
 * Both are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 * This code does not check for any "reordering" in the writes
 * corresponding to the same iteration in the reads of both dep1
 * and dep2.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we first construct
 *
 *	n_write n_read	    extra1  n_read  extra2  nparam+1
 *	z   	read1   	    read2	
 *	I   	R1	    E1      -R2	    -E2	    C1-C2   =>  z = write1 - write2
 *	0   	R3	    E3	    0	    0	    C3
 *	0   	0	    0	    R4	    E4	    C4
 *	0   	0	    0	    0	    0	    C5
 *
 * and add constraints that express that read1 is (strictly) lexicosmaller
 * than read2.
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * write1 lexicolarger than write2.
 */
static int is_reordering_pair(Polyhedron *dep1, Polyhedron *dep2,
			      int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int i, l;
    int	reordering = 0;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_write; ++i)
	value_set_si(domain->p[i][1+i], 1);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write, n_read + extra1, 1+n_write);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write, n_read + extra2, 1+n_write+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_write);

    for (l = 0; l < n_read && !reordering; ++l) {
	var_equal(domain, 1+n_write, l, 1+n_write+n_read+extra1, row);
	domain->NbRows = row + l;
	var_smaller(domain, 1+n_write+l, 1+n_write+n_read+extra1+l, row+l);
	domain->NbRows++;

#ifdef PIP_DEBUG
	pip_matrix_print(stderr, domain);
	pip_matrix_print(stderr, ctx);
#endif

	sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
	pip_quast_print(stderr, sol, 0);
#endif

	if (sol->newparm || (sol->list && lex_sign(sol->list, n_write) > 0))
	    reordering = 1;

	pip_quast_free(sol);
    }
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return reordering;
}

/* For any two pairs of corresponding writes and reads,
 * do the writes occur in a different order than the reads ?
 * If so, then the dependence is "reordering".
 */
int pip_is_reordering(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (is_reordering_pair(P1, P2, n_write, n_read, ctx))
		return 1;
    return 0;
}

/*
 * Check whether the writes from dep1 and dep2 corresponding
 * to the same read iteration occur in order.
 * Both are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we construct
 *
 *	n_write n_read	    extra1  extra2  nparam+1
 *	z   	read
 *	I   	R1-R2	    E1      -E2	    C1-C2   =>  z = write1 - write2
 *	0   	R3	    E3	    0	    C3
 *	0   	R4	    0	    E4	    C4
 *	0   	0	    0	    0	    C5
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * write1 lexicolarger than write2.
 */
static int is_inorder_accesses_pair(Polyhedron *dep1, Polyhedron *dep2,
				    int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write - n_read - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int i, l;
    int	inorder = 1;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_write; ++i)
	value_set_si(domain->p[i][1+i], 1);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write, n_read + extra1, 1+n_write);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write, n_read, 1+n_write);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read, extra2, 1+n_write+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_write);

#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    if (sol->newparm || (sol->list && lex_sign(sol->list, n_write) > 0))
	inorder = 0;

    pip_quast_free(sol);
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return inorder;
}

/* For reads from dep1 and dep2 from the same iteration,
 * does the corresponding write from dep1 occur before the write from dep2 ?
 */
int pip_is_inorder_accesses(Polyhedron *dep1, Polyhedron *dep2,
			    int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep1; P1; P1 = P1->next)
	for (P2 = dep2; P2; P2 = P2->next)
	    if (!is_inorder_accesses_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/* Check whether any read of dep1 occurs between a pair of write
 * and read of dep2
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we first construct
 *
 *	n_write2    n_read  extra1  n_read  extra2  nparam+1
 *	z   	    read1   	    read2	
 *	I	    -I	    0	    R2	    E2	    C2	    =>  z = read1 - write2
 *	0   	    R3	    E3	    0	    0	    C3
 *	0   	    0	    0	    R4	    E4	    C4
 *	0   	    0	    0	    0	    0	    C5
 *
 * and add constraints that express that read1 is lexicosmaller than or equal to
 * (in case dep1 != dep2) read2.
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * read1 lexicolarger than write2.
 */
static int is_consecutive_pair(Polyhedron *dep1, Polyhedron *dep2,
			       int n_write1, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int n_write2 = n_read;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write1 - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write1 + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write1 - n_read - nparam;
    int extra2 = dep2->Dimension - n_write2 - n_read - nparam;
    int i, l;
    int	consecutive = 1;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write1);
    assert_dependent_mapping(dep2, n_write2);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    assert(n_write2 == n_read);
    for (i = 0; i < n_write2; ++i) {
	value_set_si(domain->p[i][1+i], 1);
	value_set_si(domain->p[i][1+n_write2+i], -1);
    }
    sub_poly2pipc(dep2, domain, 0, n_write2, 0,
		  1+n_write2, n_read + extra2, 1+n_write2+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write2, 0,
		  1+n_write2+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write2;
    row = copy_domain_constraints(dep1, dep2, n_write1, n_write2, n_read, 
				  C, domain, row, n_write2);

    for (l = 0; l <= n_read && consecutive; ++l) {
	if (l == n_read && dep1 == dep2)
	    break;
	var_equal(domain, 1+n_write2, l, 1+n_write2+n_read+extra1, row);
	domain->NbRows = row + l;
	if (l < n_read) {
	    var_smaller(domain, 1+n_write2+l, 1+n_write2+n_read+extra1+l, row+l);
	    domain->NbRows++;
	}

#ifdef PIP_DEBUG
	pip_matrix_print(stderr, domain);
	pip_matrix_print(stderr, ctx);
#endif

	sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
	pip_quast_print(stderr, sol, 0);
#endif

	if (sol->newparm || (sol->list && lex_sign(sol->list, n_write2) > 0))
	    consecutive = 0;

	pip_quast_free(sol);
    }
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return consecutive;
}

/* For any pair of corresponding write and read,
 * is there any other read in between ?
 * If so, then the loop is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int pip_is_consecutive_loop(Polyhedron *dep, int n_write, int n_read, 
			    Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    assert(n_write == n_read);
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (!is_consecutive_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/* For any pair of corresponding write and read of dep2,
 * is there any read from dep1 in between ?
 * If so, then dep2 is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int pip_is_consecutive_dep(Polyhedron *dep2, Polyhedron *dep1, 
			   int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep1; P1; P1 = P1->next)
	for (P2 = dep2; P2; P2 = P2->next)
	    if (!is_consecutive_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/*
 * Check whether there is a pair of distinct reads, one from dep1 and dep2,
 * that correspond to the same write.
 * Both dependences are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we construct
 *
 *	n_read 	n_read	    extra1  n_read  extra2  nparam+1
 *	z   	read1   	    read2	
 *	I   	-I	    0	    I	    0	    0		=>  z = read1 - read2
 *	0   	R1	    E1	    -R2	    -E2	    C1-C2	=>  write1 = write2
 *	0   	R3	    E3	    0	    0	    C3
 *	0   	0	    0	    R4	    E4	    C4
 *	0   	0	    0	    0	    0	    C5
 *
 * If the lexicomaximal z is lexicopositive then such distinct reads exist.
 * Note that if the lexicomaximal z is lexiconegative, then such reads also
 * exist, but we choose to return false and this case.  Calling the function
 * again with dep1 and dep2 interchanged will return true.
 * If the lexicomaximal z is zero and if dep1 != dep2, then such reads
 * _may_ exist.  Again we return false in this case since calling the function
 * with dep1 and dep2 interchanged will determine whether they exist.
 */
static int has_multiplicity_pair(Polyhedron *dep1, Polyhedron *dep2,
			         int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int ncol = 1 + n_read + n_read + extra1 + n_read + extra2 + nparam + 1;
    int i;
    int	multiplicity = 0;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_read; ++i) {
	value_set_si(domain->p[i][1+i], 1);
	value_set_si(domain->p[i][1+n_read+i], -1);
	value_set_si(domain->p[i][1+n_read+n_read+extra1+i], 1);
    }
    row += n_read;
    poly2pipc(dep1, domain, 0, n_write, row,
	      1+n_write, n_read + extra1, 1+n_read);
    poly2pipc(dep1, domain, 0, n_write, row,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, row,
		  1+n_write, n_read + extra2, 1+n_read+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, row,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_read);

#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    if (sol->newparm || (sol->list && lex_sign(sol->list, n_read) > 0))
	multiplicity = 1;

    pip_quast_free(sol);
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return multiplicity;
}

int pip_has_multiplicity(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (has_multiplicity_pair(P1, P2, n_write, n_read, ctx))
		return 1;
    return 0;
}

Polyhedron *pip_min_ddv(Polyhedron *dep, Polyhedron *sw, Polyhedron *sr,
			int nwrite, int nread, int dim, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    Polyhedron *R;
    int nparam = C->Dimension;
    int nrow = dim + dep->NbConstraints;
    int d_extra = dep->Dimension - nwrite - nread - nparam;
    int sw_extra = sw->Dimension - dim - nwrite - nparam;
    int sr_extra = sr->Dimension - dim - nread - nparam;
    int ncol = 2 + dim + dep->Dimension + sw_extra + sr_extra;
    int i;
    int row;
    assert_dependent_mapping(sw, dim);
    assert_dependent_mapping(sr, dim);
    ctx = poly2pip(C, 0, 0, C->Dimension);

    domain = pip_matrix_alloc(nrow, ncol);
    row = 0;
    for (i = 0; i < dim; ++i)
	value_set_si(domain->p[i][1+i], 1);
    poly2pipc(sr, domain, 0, dim, 0,
	      1+dim, nread + sr_extra, 1+dim+nwrite+sw_extra);
    poly2pipc(sr, domain, 0, dim, 0,
	      1+dim+nread+sr_extra, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(sw, domain, 0, dim, 0,
		  1+dim, nwrite + sw_extra, 1+dim);
    sub_poly2pipc(sw, domain, 0, dim, 0,
		  1+dim+nwrite+sw_extra, nparam+1, domain->NbColumns-nparam-1);
    row += dim;
    poly2pipc(dep, domain, 0, dep->NbConstraints, row, 0, 1, 0);
    poly2pipc(dep, domain, 0, dep->NbConstraints, row, 1, nwrite, 1+dim);
    poly2pipc(dep, domain, 0, dep->NbConstraints, row, 
	      1+nwrite, nread, 1+dim+nwrite+sw_extra);
    poly2pipc(dep, domain, 0, dep->NbConstraints, row, 
	  1+nwrite+nread, d_extra+nparam+1, 1+dim+nwrite+sw_extra+nread+sr_extra);

#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    options = pip_options_init();
    options->Simplify = 1;
    options->Urs_unknowns = -1;
    sol = pip_solve(domain, ctx, -1, options);

#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    R = quast2poly(sol, dim, 0, nparam, C, NULL);

    pip_quast_free(sol);
    pip_options_free(options);
    pip_matrix_free(domain);
    pip_matrix_free(ctx);

    return R;
}
