#ifndef YAML_H
#define YAML_H

#include <map>
#include <string>
#include <typeinfo>
#include <vector>
#define partition algorithm_partition
#include <algorithm>
#undef partition

#include <assert.h>
#include <stdio.h>

using std::type_info;
using std::map;
using std::pair;
using std::string;

extern "C" {
#include <syck.h>
#include "yll.h"
}

#define YAML_PTR_FIELD(sd, class, name, type) do { \
    assert(typeid(&class::name) == typeid(type *class::*)); \
    sd.m[ #name ] = new ptr_field((ptr_field_t) &class::name, &typeid(type)); \
    } while(0)
#define YAML_INT_FIELD(sd, class, name) \
    sd.m[ #name ] = new int_field(static_cast<int_field_t>(&class::name), &typeid(int));
#define YAML_ENUM_FIELD(sd, class, name, names) \
    sd.m[ #name ] = new_enum_field(&class::name, \
	names, sizeof(names)/sizeof(*names));
#define YAML_SEQ_FIELD(sd, class, name, type) \
    sd.m[ #name ] = new seq_field<type> \
	    ((seq_field<type>::seq_field_t) &class::name);

namespace yaml
{

struct at_init {
    at_init(void (*f)(void)) { f(); }
};

class serialize;
class anchor_map;
typedef serialize* serialize_p;

struct description {
    serialize_p (*create) ();
    serialize_p (*convert) (anchor_map *, serialize *, const type_info *);
    const type_info	*type;
};

struct ptr_field;
struct int_field;
template <class el> struct seq_field;

class replacer_base {
protected:
    friend bool operator==(const replacer_base& a, const replacer_base& b);
    serialize*&	    ref;
public:
    replacer_base(serialize*& r) : ref(r) {}
    virtual void with(serialize *replacement) {
	ref = replacement;
    }
};

inline bool operator==(const replacer_base& a, const replacer_base& b)
{
    return &a.ref == &b.ref;
}

template <class el>
class replacer : public replacer_base {
public:
    replacer(el*& r) : replacer_base((serialize *&)r) {}
    virtual void with(serialize *replacement) {
	ref = dynamic_cast<el*>(replacement);
    }
};

typedef map<serialize *, unsigned> refset;

struct emitter {
    YllEmitter *e;
    refset	refs;
    map<serialize *, SYMID> anchors;
    emitter() { e = yll_new_emitter(); }
    ~emitter() { yll_free_emitter(e); }
    bool anchor_handled(serialize *s);
};

class serialize {
    static void syck_output_handler(YllEmitter *p, char *s, long len);
    static SYMID syck_handler(SyckParser *p, SyckNode *n);
    static SyckNode * bad_anchor_handler(SyckParser *, char *);

protected:
    friend struct ptr_field;
    friend struct int_field;
    template <class el> friend struct seq_field;

    static map<string, description*>& s_desc() {
	static map<string, description*> desc;
	return desc;
    }
    static map<const type_info *, description*>& t_desc() {
	static map<const type_info *, description*> desc;
	return desc;
    }

    static void register_type(const char *, const type_info *, description *);

public:
    static serialize * Load(SyckParser *parser, const type_info *type);
    static serialize_p Load(char *str, const type_info *type);
    static serialize_p Load(FILE *fp, const type_info *type);

    void Dump(FILE *out = stdout);
    virtual void Dump(emitter& e) {
	if (!e.anchor_handled(this))
	    dump(e);
    }
    virtual void dump(emitter& e) = 0;

    virtual void collect_references(refset& refs);

    virtual void free();

    virtual void init(SyckParser *p, SyckNode *n) = 0;

    virtual ~serialize() {
    }
};


typedef pair<SYMID, std::vector<replacer_base *> > rep_pair;

class anchor_map {
    map<serialize *, rep_pair > anchors;
    SyckParser *p;

public:
    void add(serialize* s, SYMID id) { 
	anchors[s] = rep_pair(id, std::vector<replacer_base *>());
    }
    void replace(serialize *s, replacer_base *orig, replacer_base *replace) {
	bool found = false;
	for (int i = 0; i < anchors[s].second.size(); ++i) {
	    replacer_base * r = anchors[s].second[i];
	    if (*r == *orig) {
		delete r;
		anchors[s].second[i] = replace;
		found = true;
	    }
	}
	delete orig;
	if (!found)
	    this->replace(s, replace);
    }
    void replace(serialize *s, replacer_base *replace) {
	anchors[s].second.push_back(replace);
    }
    bool is_anchor(serialize *s) { return anchors.find(s) != anchors.end(); }
    serialize * replace(SYMID id, serialize *replacement);
    serialize * replace(serialize * orig, serialize *replacement) {
	return replace(anchors[orig].first, replacement);
    }
    ~anchor_map();
    anchor_map(SyckParser *_p) : p(_p) {}
};

class bad_anchor : public serialize {
public:
    void init(SyckParser *p, SyckNode *n) {};
    void dump(emitter& e) { assert(0); }
};

class str : public serialize {
    str() {}
public:
    string	s;
    str(string _s) : s(_s) {}
    static void register_type();
    static serialize_p create();
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e) { yll_emitter_write_string(e.e, s.c_str()); }
};

class integer : public serialize {
    static serialize_p convert (anchor_map *, serialize *, const type_info *);

    integer() {}
public:
    int		v;
    integer(int _v) : v(_v) {}
    static void register_type();
    static serialize_p create();
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e) { yll_emitter_write_int(e.e, v); }
};

typedef map<string, serialize *>  sermap;

typedef map<string,serialize_p>::mapped_type mapped_type;

class dict : public serialize {
protected:

public:
    map<string,serialize_p>  *d;

    //static void register_type();
    static serialize_p create();
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e);

    void collect_references(refset& refs);

    mapped_type& operator[](string s) { return (*d)[s]; }

    ~dict() {
	if (d)
	    delete d;
    }
};

template <class el>
class seq : public serialize {
public:
    std::vector<el *>  v;

    //static void register_type();
    static serialize_p create() { return new seq<el>; }

    void init(SyckParser *p, SyckNode *n) {
	anchor_map *anchors = (anchor_map *) p->bonus;
	for (int i = 0; i < n->data.list->idx; i++) {
	    el *val;
	    SYMID oid = syck_seq_read(n, i);
	    syck_lookup_sym( p, oid, (char **)&val );
	    v.push_back(val);
	}
	// In a second iteration, since we wouldn't want the
	// memory locations to change for the replacer
	for (int i = 0; i < n->data.list->idx; i++) {
	    if (anchors->is_anchor(v[i])) {
		anchors->replace(v[i], new replacer<el>(v[i]));
	    }
	}
    }

    void collect_references(refset& refs) {
	serialize::collect_references(refs);
	for (int i = 0; i < v.size(); ++i) {
	    el *e = v[i];
	    if (refs.find(e) == refs.end())
		e->collect_references(refs);
	    else
		++refs[e];
	}
    }

    void dump(emitter& e) {
	yll_emitter_start_seq(e.e);
	for (int i = 0; i < v.size(); ++i) {
	    v[i]->Dump(e);
	}
	yll_emitter_end_seq(e.e);
    }

    el*& operator[](int i) { return v[i]; }
    el* const & operator[](int i) const { return v[i]; }
    size_t size() const { return v.size(); }
    void push_back(el *e) { v.push_back(e); }
    void resize(int n) { v.resize(n); }
    void erase(el *e) { if (find(v.begin(), v.end(), e) != v.end())
			    v.erase(find(v.begin(), v.end(), e)); }

    ~seq() {
    }
};

template <>
class seq<int> : public serialize {
public:
    std::vector<int>  v;

    //static void register_type();
    static serialize_p create() { return new seq<int>; }

    void init(SyckParser *p, SyckNode *n) {
	for (int i = 0; i < n->data.list->idx; i++) {
	    str* val;
	    SYMID oid = syck_seq_read(n, i);
	    syck_lookup_sym( p, oid, (char **)&val );
	    v.push_back(atoi(val->s.c_str()));
	}
    }

    void collect_references(refset& refs) {
	serialize::collect_references(refs);
    }

    void dump(emitter& e) { assert(0); }

    int& operator[](int i) { return v[i]; }
    int const & operator[](int i) const { return v[i]; }
    size_t size() const { return v.size(); }
    void push_back(int e) { v.push_back(e); }
    void resize(int n) { v.resize(n); }

    ~seq() {
    }
};

class structure;
typedef serialize * structure::* 	ptr_field_t;
typedef int structure::* 		int_field_t;

struct field {
    const type_info		*type;
    field() {};
    field(const type_info *t) : type(t) {};
    virtual bool set(anchor_map *am,  structure *, serialize *&) { assert(0); }
    virtual void collect(structure * s, refset& refs) { assert(0); }
    virtual void dump(structure * s, emitter& refs) { assert(0); }
    virtual bool is_empty(structure *) { return false; }
};

struct ptr_field : field {
    serialize * structure::* 	pos;
    ptr_field(ptr_field_t p, const type_info *t) : pos(p), field(t) {};
    bool set(anchor_map *am, structure *, serialize *&);
    void collect(structure * s, refset& refs);
    void dump(structure * s, emitter& refs);
    bool is_empty(structure * s) { return !(s->*pos); }
};

struct int_field : field {
    int structure::* 	pos;
    int_field(int_field_t p, const type_info *t) : pos(p), field(t) {};
    bool set(anchor_map *am, structure *, serialize *&);
    void collect(structure * s, refset& refs);
    void dump(structure * s, emitter& refs);
};

template <typename T, class C>
struct enum_field : field {
    T C::* 	pos;
    const char **names;
    const int size;
    enum_field(T C::* p, const char **names, 
	const int size) : names(names), size(size), pos(p), field(&typeid(T)) {};
    bool set(anchor_map *am, structure *, serialize *&);
    void collect(structure * s, refset& refs) { }
    void dump(structure * s, emitter& refs);
};

template <typename T, class C>
bool enum_field<T,C>::set(anchor_map *am, structure * sc, serialize *& el)
{
    str *s = dynamic_cast<str *>(el);
    C *c = dynamic_cast<C *>(sc);
    if (!s)
	return false;

    for (int i = 0; i < size; ++i)
	if (!strcmp(s->s.c_str(), names[i])) {
	    c->*pos = static_cast<T>(i);
	    delete el;
	    return true;
	}

    return false;
}

template <typename T, class C>
void enum_field<T,C>::dump(structure * sc, emitter& e)
{
    C *c = dynamic_cast<C *>(sc);
    if (c->*pos >= 0 && c->*pos < size)
	yll_emitter_write_string(e.e, names[c->*pos]);
    else
	yll_emitter_write_null(e.e);
}

template <typename T, class C>
inline field *new_enum_field(T C::* p, const char **names, const int size)
{
    return new enum_field<T,C>(p, names, size);
}

template <class el>
struct seq_field : field {
    typedef seq<el> structure::* 		seq_field_t;

    seq<el> structure::* 	pos;
    seq_field(seq_field_t p) : pos(p), field(&typeid(el)) {};
    bool set(anchor_map *am, structure * c, serialize * &s) {
	map<const type_info *, description*>& desc = serialize::t_desc();
	seq<serialize> * sq = dynamic_cast<seq<serialize>*>(s);
	if (!sq)
	    return false;
	(c->*pos).v = std::vector<el *>(sq->size());
	for (int i = 0; i < sq->size(); ++i) {
	    serialize * e = sq->v[i];
	    if (typeid(el) != typeid(*e))
		e = desc[&typeid(el)]->convert(am, e, type);

	    (c->*pos).v[i] = dynamic_cast<el *>(e);

	    if (am->is_anchor(e)) {
		am->replace(e, new replacer<serialize>(sq->v[i]),
			       new replacer<el>((c->*pos).v[i]));
	    }
	}
	delete sq;
	return true;
    }
    void collect(structure * c, refset& refs) {
	std::vector<el *> & v = (c->*pos).v;
	for (int i = 0; i < v.size(); ++i) {
	    serialize *e = v[i];
	    if (!e)
		continue;
	    if (refs.find(e) == refs.end())
		e->collect_references(refs);
	    else
		++refs[e];
	}
    }
    void dump(structure * s, emitter& emit) {
	std::vector<el *> & v = (s->*pos).v;
	yll_emitter_start_seq(emit.e);
	for (int i = 0; i < v.size(); ++i) {
	    serialize *e = v[i];
	    if (e)
		e->Dump(emit);
	    else
		yll_emitter_write_null(emit.e);
	}
	yll_emitter_end_seq(emit.e);
    }
    bool is_empty(structure * s) { return (s->*pos).v.size() == 0; }
};

template <>
struct seq_field<int> : field {
    typedef seq<int> structure::* 		seq_field_t;

    seq<int> structure::* 	pos;
    seq_field(seq_field_t p) : pos(p), field(&typeid(int)) {};
    bool set(anchor_map *am, structure * c, serialize * &s) {
	map<const type_info *, description*>& desc = serialize::t_desc();
	seq<serialize> * sq = dynamic_cast<seq<serialize>*>(s);
	if (!sq)
	    return false;
	(c->*pos).v = std::vector<int>(sq->size());
	for (int i = 0; i < sq->size(); ++i) {
	    serialize * e = sq->v[i];
	    if (typeid(str) != typeid(*e))
		e = desc[&typeid(str)]->convert(am, e, type);
	    str * s = dynamic_cast<str *>(e);

	    (c->*pos).v[i] = atoi(s->s.c_str());
	    delete s;
	}
	delete sq;
	return true;
    }
    void collect(structure * c, refset& refs) {
    }
    void dump(structure * s, emitter& emit) {
	std::vector<int> & v = (s->*pos).v;
	yll_emitter_set_options(emit.e, YLL_INLINE_SEQ);
	yll_emitter_start_seq(emit.e);
	for (int i = 0; i < v.size(); ++i) {
	    yll_emitter_write_int(emit.e, v[i]);
	}
	yll_emitter_end_seq(emit.e);
    }
};

typedef map<string, field *>  fieldmap;

struct struct_description {
    description	    d;
    fieldmap	    m;
};

class structure : public serialize {
    static serialize_p convert (anchor_map *, serialize *, const type_info *);

public:
    map<string,serialize_p>  *d;

    static void register_type(const char *, const type_info *, description *);

    void collect_references(refset& refs);
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e);
    virtual void fill_fields(anchor_map *);

    structure() { d = NULL; }
    ~structure() { if (d) delete d; }
};

serialize_p Load(char *str, const type_info *type = NULL);
serialize_p Load(FILE *fp, const type_info *type = NULL);
template <class T>
T* Load(char *str) {
    return dynamic_cast<T*>(Load(str, &typeid(T)));
}
template <class T>
T* Load(FILE *fp) {
    return dynamic_cast<T*>(Load(fp, &typeid(T)));
}

}

#endif
