#include <stdlib.h>
#include "yll.h"

#define FL_INIT(l, f)   (l) = (f)               /* Specific flags location. */
#define FL_SET(l, f)    ((l) |= (f))
#define FL_CLR(l, f)    ((l) &= ~(f))
#define FL_ISSET(l, f)  ((l) & (f))

#define F_INIT(p, f)    FL_INIT((p)->flags, f)  /* Structure element flags. */
#define F_SET(p, f)     FL_SET((p)->flags, f)
#define F_CLR(p, f)     FL_CLR((p)->flags, f)
#define F_ISSET(p, f)   FL_ISSET((p)->flags, f)

static void yll_emit_space(YllEmitter *e)
{
    if (F_ISSET(e->stack, YLL_ADD_SPACE))
	fprintf(e->out, " ");
    F_CLR(e->stack, YLL_ADD_SPACE);
}

static void yll_emit_transfer(YllEmitter *e)
{
    yll_emit_space(e);
    if (e->transfer) {
	fprintf(e->out, "!%s", e->transfer);
	e->transfer = 0;
	F_SET(e->stack, YLL_ADD_SPACE);
    }
}

static void yll_push(YllEmitter *e, YllEmitterNode *n)
{
    n->next = e->stack;
    e->stack = n;
    n->flags = e->flags;
    e->flags = 0;
    e->indent += 2;
}

static void yll_pop(YllEmitter *e)
{
    YllEmitterNode *n = e->stack;
    e->stack = e->stack->next;
    S_FREE(n);
    e->indent -= 2;
}

static void next_doc(YllEmitter *e)
{
    fprintf(e->out, "--- ");
}

static void next_anchor(YllEmitter *e)
{
    YllAnchorNode *n = (YllAnchorNode*) e->stack;
    SYMID a = n->anchor;
    yll_pop(e);
    e->stack->nextf(e);
    fprintf(e->out, "&%d", a);
    F_SET(e->stack, YLL_ADD_SPACE);
}

static void next_seq(YllEmitter *e)
{
    YllSeqNode *n = (YllSeqNode*) e->stack;
    if (F_ISSET(e->stack, YLL_INLINE_SEQ)) {
	if (!n->first)
	    fprintf(e->out, ", ");
    } else
	fprintf(e->out, "\n%*s- ", e->indent, "");
    n->first = 0;
}

static void next_map_val(YllEmitter *e);

static void next_map_key(YllEmitter *e)
{
    F_CLR(e->stack, YLL_ADD_SPACE);
    fprintf(e->out, "\n%*s", e->indent, "");
    e->stack->nextf = next_map_val;
}

static void next_map_val(YllEmitter *e)
{
    fprintf(e->out, ": ");
    e->stack->nextf = next_map_key;
}

YllEmitter* yll_new_emitter() {
    YllEmitter *e;
    YllEmitterNode *n;
    e = S_ALLOC(YllEmitter);
    n = S_ALLOC(YllEmitterNode);
    n->nextf = next_doc;
    e->last_anchor = 0;
    e->stack = 0;
    e->flags = 0;
    e->transfer = 0;
    yll_push(e, n);
    e->indent = -2;
    return e;
}

void yll_emitter_handler(YllEmitter *e, YllOutputHandler hdlr) 
{
}

void yll_free_emitter(YllEmitter *e) 
{
    fprintf(e->out, "\n");
    yll_pop(e);
    S_FREE(e);
}

void yll_emitter_start_seq(YllEmitter *e)
{
    YllSeqNode * n;
    e->stack->nextf(e);
    n = S_ALLOC(YllSeqNode);
    n->en.nextf = next_seq;
    n->first = 1;
    yll_push(e, (YllEmitterNode *)n);
    yll_emit_transfer(e);
    if (F_ISSET(&n->en, YLL_INLINE_SEQ))
	fprintf(e->out, "[");
}

void yll_emitter_end_seq(YllEmitter *e)
{
    YllSeqNode *n = (YllSeqNode*) e->stack;
    if (F_ISSET(e->stack, YLL_INLINE_SEQ))
	fprintf(e->out, "]");
    else if (n->first) {
	if (F_ISSET(e->stack, YLL_ADD_SPACE))
	    fprintf(e->out, " ");
	fprintf(e->out, "[]");
    }
    yll_pop(e);
}

void yll_emitter_start_map(YllEmitter *e)
{
    YllEmitterNode * n;
    e->stack->nextf(e);
    n = S_ALLOC(YllEmitterNode);
    n->nextf = next_map_key;
    yll_push(e, n);
    yll_emit_transfer(e);
}

void yll_emitter_end_map(YllEmitter *e)
{
    yll_pop(e);
}

SYMID yll_emitter_anchor(YllEmitter *e)
{
    YllAnchorNode * n;
    SYMID a = ++e->last_anchor;
    n = S_ALLOC(YllAnchorNode);
    n->anchor = a;
    n->en.nextf = next_anchor;
    yll_push(e, (YllEmitterNode *)n);
    return a;
}

void yll_emitter_alias(YllEmitter *e, SYMID id)
{
    e->stack->nextf(e);
    fprintf(e->out, "*%d", id);
}

void yll_emitter_write_string(YllEmitter *e, const char *str)
{
    e->stack->nextf(e);
    yll_emit_transfer(e);
    yll_emit_space(e);
    if (strchr(str, ':'))
	fputc('\'', e->out);
    fprintf(e->out, "%s", str);
    if (strchr(str, ':'))
	fputc('\'', e->out);
}

void yll_emitter_write_int(YllEmitter *e, int i)
{
    e->stack->nextf(e);
    yll_emit_transfer(e);
    yll_emit_space(e);
    fprintf(e->out, "%d", i);
}

void yll_emitter_write_null(YllEmitter *e)
{
    e->stack->nextf(e);
    yll_emit_transfer(e);
    yll_emit_space(e);
    fprintf(e->out, "~");
}

void yll_emitter_set_options(YllEmitter *e, YllOption o)
{
    F_SET(e, o);
}

void yll_emitter_set_transfer(YllEmitter *e, const char *t)
{
    e->transfer = t;
}
